'''Converters between vote and result formats.

These objects have a `convert()` method that converts between different formats
of votes or election results.
'''

import collections
import statistics
import builtins
import operator
import itertools
import decimal
from fractions import Fraction
from decimal import Decimal
from typing import Any, List, Tuple, Dict, FrozenSet, Union, Callable, \
                   Iterable, Collection
from numbers import Number

import votelib.candidate
import votelib.util
import votelib.vote
from votelib.candidate import \
    Candidate, Constituency, IndividualElectionOption, ElectionParty, \
    IndividualToPartyMapper
from votelib.vote import RankedVoteType, ScoreVoteType, VoteSubsetter
from votelib.component import rankscore
from votelib.persist import simple_serialization


def _subtract_lowest(scores: Dict[Any, int],
                     sorted_scores: List[Any],
                     cutoff: int,
                     ) -> None:
    cut = 0
    for score in sorted_scores:
        if scores[score] <= (cutoff - cut):
            cut += scores[score]
            del scores[score]
        else:
            scores[score] -= (cutoff - cut)
            break


class Converter:
    def convert(self, *args, **kwargs):
        raise NotImplementedError


# Aggregators from more complicated votes to simple magnitudes
@simple_serialization
class ApprovalToSimpleVotes:
    '''Aggregate approval votes to simple votes.

    Aggregate votes for candidate sets (approval votes) to separate votes
    for individual candidates. Useful for example in approval voting.

    :param split: Whether to split the vote power among all the candidates
        in the set (as in satisfaction approval voting) [#wsav]_
        or give full vote to each (as in ordinary approval voting) [#wav]_.

    .. [#wsav] "Satisfaction approval voting", Wikipedia.
        https://en.wikipedia.org/wiki/Satisfaction_approval_voting

    .. [#wav] "Approval voting", Wikipedia.
        https://en.wikipedia.org/wiki/Approval_voting
    '''
    def __init__(self, split: bool = False):
        self.split = split

    def convert(self,
                votes: Dict[FrozenSet[Candidate], int]
                ) -> Dict[Candidate, Number]:
        '''Convert approval votes to simple votes.'''
        agg_votes = collections.defaultdict(int)
        for bulk, n_votes in votes.items():
            if self.split:
                n_votes = Fraction(n_votes, len(bulk))
            for cand in bulk:
                agg_votes[cand] += n_votes
        return dict(agg_votes)


@simple_serialization
class ScoreToSimpleVotes:
    '''Aggregate scores (cardinal votes) to simple votes.

    Useful for score voting (range voting) systems. Note that, if the usual
    central value functions such as mean or median are used, this does not
    give scores that could be passed to ordinary magnitude-based evaluators
    since mean/median scores do not scale with the number of voters.

    This can be used directly in combination with a simple-vote plurality
    evaluator but is also a component in several cardinal vote evaluators.

    The scores can be arbitrary values as long as the aggregation function
    can cope with them, but numeric scores are the most common.

    :param function: A function to aggregate all scores given to a candidate
        to a single score. It can also be specified by name (string); this
        looks into the exact aggregator register in the utility module (which
        reimplements some common aggregation functions to use exact arithmetic,
        such as the mean, which is the default here) and then searches builtin
        functions and the statistics stdlib module namespaces.
    :param unscored_value: Score to give to a candidate that was not assigned
        a score by the voter. None means such ballots will not be considered
        for the candidate. A function can be specified to determine the value
        from all assigned scores (such as assigning the minimum). Builtin
        functions and functions from the statistics stdlib module can be
        specified by their names.
    :param min_count: Minimum count of voter scores for the candidate to be
        considered. Candidates below this threshold will be assigned
        bottom_value. This is used in some systems to prevent an unknown
        upset candidate to win by high score mean although they were scored
        only by a handful of highly dedicated voters.
    :param truncation: Fraction (if lower than 1) or count (if at least 1)
        of lowest and highest scores to disregard before aggregating, to
        stabilize the result. (Both ends are trimmed using this, so the
        number/fraction of scores disregarded is twice the count/fraction.)
    :param bottom_value: Value to assign to candidates with less voter scores
        than min_count. This would usually be the lowest possible aggregate
        score.
    '''
    FUNCTION_NAMESPACES = [
        statistics,
        builtins,
    ]

    def __init__(self,
                 function: Union[
                     Callable[[List[Any]], Any], str
                 ] = 'mean',
                 unscored_value: Union[
                     Callable[[List[Any]], Any], Any, None
                 ] = None,
                 min_count: int = 0,
                 truncation: Number = 0,
                 bottom_value: Any = 0,
                 ):
        self.function = self._get_function(function)
        if isinstance(unscored_value, str):
            unscored_value = self._get_function(unscored_value)
        self.unscored_value = unscored_value
        self.min_count = min_count
        self.truncation = truncation
        self.bottom_value = bottom_value

    def _get_function(self, fdef) -> Callable[[List[Number]], Number]:
        if isinstance(fdef, str):
            if fdef in votelib.util.EXACT_AGGREGATORS:
                return votelib.util.EXACT_AGGREGATORS[fdef]
            for namespace in self.FUNCTION_NAMESPACES:
                if hasattr(namespace, fdef):
                    return getattr(namespace, fdef)
            raise ValueError(f'unknown function definition: {fdef!r}')
        else:
            if not hasattr(fdef, '__call__'):
                raise ValueError(f'not a callable function: {fdef!r}')
            return fdef

    def convert(self,
                votes: Dict[ScoreVoteType, int],
                ) -> Dict[Candidate, Any]:
        '''Convert score votes to simple votes.

        :param votes: Uncorrected score votes.
        :returns: A mapping from candidates to their aggregate scores.
        '''
        return self.aggregate(self.corrected_scores(votes))

    def corrected_scores(self,
                         votes: Dict[ScoreVoteType, int],
                         ) -> Dict[Candidate, Dict[Any, int]]:
        '''Correct score votes to an intermediate format.

        This forms the first part of the conversion (the second is
        :meth:`aggregate`) and is exposed independently to allow for repeated
        aggregation which is used by some score voting evaluators.

        The correction adds values for unscored candidates, assigns a fixed
        value to candidates with too few votes, and potentially truncates the
        scores to enable truncated central values.

        :param votes: Uncorrected score votes.
        :returns: Corrected scores in a nested dictionary (candidates are
            mapped to dictionaries mapping score values to numbers of their
            occurrences).
        '''
        scores = collections.defaultdict(lambda: collections.defaultdict(int))
        for one_vote, n in votes.items():
            for cand, score in one_vote:
                scores[cand][score] += n
        n_votes = sum(votes.values())
        return {
            cand: self._correct_candidate_scores(cscores, n_votes)
            for cand, cscores in scores.items()
        }

    def aggregate(self,
                  scores: Dict[Candidate, Dict[Any, int]]
                  ) -> Dict[Candidate, Any]:
        '''Aggregate corrected score votes to scores per candidate.

        :param scores: Corrected scores in an intermediate format of a nested
            dict (candidates are mapped to dictionaries mapping score values to
            numbers of their occurrences).
        :returns: A mapping from candidates to their aggregate scores.
        '''
        return {
            cand: self.aggregate_one(cscores)
            for cand, cscores in scores.items()
        }

    def aggregate_one(self, cscores: Dict[Any, int]) -> Any:
        '''Aggregate corrected scores for a candidate to a single score.'''
        return self.function([
            score for score, count in cscores.items() for i in range(count)
        ])

    def _correct_candidate_scores(self,
                                  scores: Dict[Any, int],
                                  n_votes: int = None,
                                  ) -> Dict[Any, int]:
        n_scores = sum(scores.values())
        if n_scores < self.min_count:
            return {self.bottom_value: self.min_count}
        copied = False
        if n_votes is not None and self.unscored_value is not None:
            if hasattr(self.unscored_value, '__call__'):
                unscored = self.unscored_value([
                    score for score, count in scores.items()
                        for i in range(count)    # noqa: E131
                ])
            else:
                unscored = self.unscored_value
            if not copied:
                scores = scores.copy()
                copied = True
            scores[unscored] = n_votes - n_scores + scores.get(unscored, 0)
        if self.truncation > 0:
            if not copied:
                scores = scores.copy()
                copied = True
            if self.truncation < 1:
                cutoff = int(
                    (n_votes if n_votes else n_scores) * self.truncation
                )
            else:
                cutoff = self.truncation
            sorted_scores = list(sorted(scores.keys()))
            _subtract_lowest(scores, sorted_scores, cutoff)
            _subtract_lowest(scores, sorted_scores[::-1], cutoff)
        return scores


@simple_serialization
class RankedToFirstPreference:
    '''Aggregate ranked votes to simple votes, taking each voter's first choice.

    This is useful to determine the plurality winner in ranked choice voting
    systems.
    '''
    def convert(self,
                votes: Dict[RankedVoteType, int],
                ) -> Dict[Candidate, int]:
        '''Convert ranked votes to simple votes by taking first choices.'''
        output = collections.defaultdict(int)
        for ranking, n_votes in votes.items():
            if ranking:
                output[ranking[0]] += n_votes
        return dict(output)


@simple_serialization
class RankedToFirstNPreferences:
    '''Aggregate ranked votes to approval votes, taking first N choices.

    :param n_first: Number of top choices to take from each ballot.
    '''
    def __init__(self, n_first: int):
        self.n_first = n_first

    def convert(self,
                votes: Dict[RankedVoteType, int],
                ) -> Dict[Candidate, int]:
        output = collections.defaultdict(int)
        for ranking, n_votes in votes.items():
            if ranking:
                output[frozenset(ranking[:self.n_first])] += n_votes
        return dict(output)


@simple_serialization
class RankedToPresenceCounts:
    '''Count candidate occurrences in ranked votes regardless of rank.

    Returns simple votes. The dictionary will be ordered in the order of the
    ranked votes - first, all candidates appearing on any first rank will be
    listed in the order of their first such votes, then all candidates
    that appear on second and lower ranks only, etc.
    '''
    def convert(self,
                votes: Dict[RankedVoteType, Number],
                ) -> Dict[Candidate, Number]:
        '''Convert ranked votes to simple votes, disregarding rank.'''
        output = collections.defaultdict(int)
        for cand, rank_i, n_votes in votelib.util.all_rankings(votes):
            output[cand] += votes
        return dict(output)


@simple_serialization
class RankedToApprovalVotes:
    '''Convert ranked votes to approval votes (disregarding rank).

    Returns approval (set) votes by lumping together all candidates ranked on
    a particular ballot regardless of rank.
    '''
    def convert(self,
                votes: Dict[RankedVoteType, Number],
                ) -> Dict[FrozenSet[Candidate], Number]:
        approval = {}
        for ranking, n_votes in votes.items():
            vote_cands = set()
            for positioned in ranking:
                if isinstance(positioned, collections.abc.Set):
                    vote_cands.update(positioned)
                else:
                    vote_cands.add(positioned)
            approval[frozenset(vote_cands)] = n_votes
        return approval


@simple_serialization
class RankedToPositionalVotes:
    '''Aggregate ranked votes to simple votes.

    Useful for Borda count systems. Assigns a score to each rank and then
    sums the scores. The complete Borda count system is obtained by coupling
    this converter with the Plurality evaluator.

    :param rank_scorer: A rank scorer that determines which score to assign to
        which rank through its `scores()` method.
        You can use any object that honors the interface of
        :class:`rankscore.RankScorer`, such as any of its subclasses.
    :param unranked_scoring: What score to assign to candidates not ranked by
        the given voter. So far only the `'zero'` option, which assigns a score
        of zero, is supported.
    '''
    def __init__(self,
                 rank_scorer: rankscore.RankScorer,
                 unranked_scoring: str = 'zero',
                 ):
        self.rank_scorer = rank_scorer
        self.unranked_scoring = unranked_scoring
        if self.unranked_scoring != 'zero':
            raise NotImplementedError

    def convert(self,
                votes: Dict[RankedVoteType, int],
                ) -> Dict[Candidate, Number]:
        '''Convert ranked votes to simple votes by scoring their positions.'''
        all_candidates = votelib.util.all_ranked_candidates(votes)
        if hasattr(self.rank_scorer, 'set_n_candidates'):
            self.rank_scorer.set_n_candidates(len(all_candidates))
        rank_scores = collections.defaultdict(list)
        agg_votes = {cand: 0 for cand in all_candidates}
        for ranked, n_votes in votes.items():
            n_ranks = len(ranked)
            this_rank_scores = rank_scores.setdefault(
                n_ranks,
                self.rank_scorer.scores(n_ranks)
            )
            for rank, positioned in enumerate(ranked):
                score = this_rank_scores[rank] * n_votes
                if hasattr(positioned, '__len__'):
                    for cand in positioned:
                        agg_votes[cand] += score
                else:
                    agg_votes[positioned] += score
        return votelib.util.descending_dict(agg_votes)


@simple_serialization
class RankedToCondorcetVotes:
    '''Aggregate ranked votes to counts of pairwise wins.

    Basic component for Condorcet methods. For each ballot that ranks a pair
    of candidates in a given order, adds one to the count of the first
    candidate over the second.

    :param unranked_at_bottom: Whether to consider candidates not ranked on a
        ballot as being ranked last. If False, these candidates are not
        considered (the voter is assumed not to have any preferences there).
    '''
    def __init__(self, unranked_at_bottom: bool = True):
        self.unranked_at_bottom = unranked_at_bottom

    def convert(self,
                votes: Dict[RankedVoteType, int],
                ) -> Dict[Tuple[Candidate, Candidate], int]:
        '''Convert ranked votes to counts of pairwise wins.'''
        all_cands = frozenset(votelib.util.all_ranked_candidates(votes))
        counts = collections.defaultdict(int)
        for ranking, n_votes in votes.items():
            is_bulk = []
            ranked = []
            for item in ranking:
                is_item_bulk = isinstance(item, collections.abc.Set)
                is_bulk.append(is_item_bulk)
                if is_item_bulk:
                    ranked.extend(item)
                else:
                    ranked.append(item)
            if self.unranked_at_bottom:
                unranked = all_cands.difference(ranked)
            for i, upper_item in enumerate(ranking):
                if not is_bulk[i]:
                    upper_item = (upper_item, )
                for upper_cand in upper_item:
                    for j, lower_item in enumerate(ranking[i+1:]):
                        if not is_bulk[i+1+j]:
                            lower_item = (lower_item, )
                        for lower_cand in lower_item:
                            counts[upper_cand, lower_cand] += n_votes
                    if self.unranked_at_bottom:
                        for unranked_cand in unranked:
                            counts[upper_cand, unranked_cand] += n_votes
        return dict(counts)


@simple_serialization
class ScoreToRankedVotes:
    '''Convert score votes to ranked votes.

    This is useful to employ ranked voting methods for run-off in some score
    voting systems to reduce their susceptibility to tactical voting (e.g. STAR
    voting).

    :param unscored_value: Score to give to a candidate that was not assigned
        a score by the voter. None means such ballots will not be considered
        for the candidate. A callable is not accepted.
    '''
    def __init__(self,
                 unscored_value: Union[str, Number, None] = None,
                 ):
        if hasattr(unscored_value, '__call__'):
            raise ValueError('callables are not accepted')
        self.unscored_value = unscored_value

    def convert(self,
                votes: Dict[ScoreVoteType, int],
                ) -> Dict[RankedVoteType, int]:
        '''Convert score votes to ranked votes.

        :param votes: Score votes.
        '''
        all_candidates = frozenset(
            cand for vote, n_votes in votes.items() for cand, score in vote
        )
        output = collections.defaultdict(int)
        for scored, n_votes in votes.items():
            ranked = self.convert_one(scored, all_candidates)
            output[ranked] += n_votes
        return dict(output)

    def convert_one(self,
                    vote: ScoreVoteType,
                    all_candidates: FrozenSet[Candidate],
                    ) -> RankedVoteType:
        key_fx = operator.itemgetter(1)
        if self.unscored_value is not None:
            unscored = all_candidates.difference(cand for cand, score in vote)
            if unscored:
                vote |= ((cand, self.unscored_value) for cand in unscored)
        sorted_iter = sorted(vote, key=key_fx)
        ranked = []
        for score, tuple_iter in itertools.groupby(sorted_iter, key=key_fx):
            tuples = list(tuple_iter)
            if len(tuples) == 1:
                ranked.append(tuples[0][0])
            else:
                ranked.append(frozenset(t[0] for t in tuples))
        return tuple(reversed(ranked))


@simple_serialization
class ScoreToApprovalVotesThreshold:
    '''Convert score votes to approval votes by a score approval threshold.

    Only approves candidates whose score is numerically greater than or equal
    to the specified threshold.
    '''
    def __init__(self, threshold: Number):
        self.threshold = threshold

    def convert(self,
                votes: Dict[ScoreVoteType, int],
                ) -> Dict[FrozenSet[Candidate], int]:
        approvals = collections.defaultdict(int)
        for vote, n_votes in votes.items():
            approved = frozenset(
                cand for cand, score in vote if score >= self.threshold
            )
            if approved:
                approvals[approved] += n_votes
        return dict(approvals)


# Inverters
@simple_serialization
class InvertedSimpleVotes:
    '''Vote inverter to represent negative simple votes.

    In some voting systems, voters vote against rather than for candidates.
    '''
    def convert(self,
                votes: Dict[Candidate, Number]
                ) -> Dict[Candidate, Number]:
        '''Invert the count signs of single votes.'''
        return {cand: -n_votes for cand, n_votes in votes.items()}


# Individual/party converters
@simple_serialization
class IndividualToPartyVotes:
    '''Aggregate votes for individual candidates to votes for their parties.

    Useful for cases where votes are received by candidates but also considered
    by party, e.g. in panachage systems.

    :param mapper: A mapper object specifying the mapping from individuals to
        parties.
    '''
    DEFAULT_MAPPER = IndividualToPartyMapper()

    def __init__(self, mapper: IndividualToPartyMapper = DEFAULT_MAPPER):
        self.mapper = mapper

    def convert(self,
                votes: Dict[IndividualElectionOption, int],
                ) -> Dict[ElectionParty, int]:
        '''Convert individual simple votes to party-based simple votes.'''
        aggregated = collections.defaultdict(int)
        for cand, n in votes.items():
            party = self.mapper(cand)
            if party is not IndividualToPartyMapper.IGNORE:
                aggregated[party] += n
        return dict(aggregated)


@simple_serialization
class IndividualToPartyResult:
    '''Aggregate individual elected candidates to results for their parties.

    Useful to determine party results in systems (or parts thereof) where
    party affiliation is not taken into account during evaluation,
    e.g. in STV systems (Ireland) or in mixed-member proportional systems
    (Germany) to determine the number of directly elected candidates
    before party-based seats are allocated.

    :param mapper: A mapper object specifying the mapping from individuals to
        parties.
    '''
    DEFAULT_MAPPER = IndividualToPartyMapper()

    def __init__(self, mapper: IndividualToPartyMapper = DEFAULT_MAPPER):
        self.mapper = mapper

    def convert(self,
                results: List[IndividualElectionOption],
                ) -> Dict[ElectionParty, int]:
        '''Convert individual selection results to party-based counts.'''
        aggregated = collections.defaultdict(int)
        for cand in results:
            party = self.mapper(cand)
            if party is not IndividualToPartyMapper.IGNORE:
                aggregated[party] += 1
        return dict(aggregated)


@simple_serialization
class GroupVotesByParty:
    '''Group individual simple votes to a dict nested by their parties.

    Useful when the votes are provided for individual candidates and should be
    retained as such, but are evaluated per-party (e.g. in panachage
    open-list systems, where this can combine list votes for party list
    evaluation).

    :param mapper: A mapper object specifying the mapping from individuals to
        parties.
    '''

    DEFAULT_MAPPER = IndividualToPartyMapper()

    def __init__(self, mapper: IndividualToPartyMapper = DEFAULT_MAPPER):
        self.mapper = mapper

    def convert(self,
                votes: Dict[IndividualElectionOption, int],
                ) -> Dict[ElectionParty, Dict[IndividualElectionOption, int]]:
        '''Group individual simple votes by individuals' parties.'''
        aggregated = {}
        for cand, n in votes.items():
            party = self.mapper(cand)
            if party is not IndividualToPartyMapper.IGNORE:
                aggregated.setdefault(party, {})[cand] = n
        return aggregated


@simple_serialization
class SelectionToDistribution:
    '''Adapt selection results to distribution (seat count) format.

    This can be used e.g. for majority bonus systems where the largest party
    gets a predetermined amount of additional reserved seats, or in
    mixed-member proportional systems to determine party-wise results of the
    constituency round.

    :param amount: How many votes to attribute to each winner of the selection.
    '''
    def __init__(self, amount: Number = 1):
        self.amount = amount

    def convert(self, elected: List[Candidate]) -> Dict[Candidate, int]:
        '''Convert selection results to distribution results.'''
        return {cand: self.amount for cand in elected}


@simple_serialization
class MergedSelections:
    '''Compile candidates elected in constituencies to a single result list.

    The candidates are ordered by their positions in the district-wide result
    lists.

    Useful e.g. in mixed-member proportional systems to list all candidates
    elected in constituencies.
    '''
    def convert(self, elected: Union[
                    Dict[Constituency, List[Candidate]],
                    List[List[Candidate]]
                ]) -> List[Candidate]:
        '''Compile constituency election results to a single result list.

        :param elected: Partial selection results in a list or dictionary;
            if a dictionary, its keys are ignored and values treated as a list.
        '''
        # TODO: reconcile ties when merging?
        if hasattr(elected, 'values'):
            elected = elected.values()
        ranks = self._get_ranks(elected)
        return list(sorted(
            ranks.keys(),
            key=(lambda cand: (-len(ranks[cand]), -sum(ranks[cand])))
        ))

    @staticmethod
    def _get_ranks(elected: Iterable[List[Candidate]]):
        ranks = {}
        for clist in elected:
            max_rank = len(clist) - 1
            for i, cand in enumerate(clist):
                if cand not in ranks:
                    ranks[cand] = []
                ranks[cand].append(max_rank - i)
        return ranks


@simple_serialization
class MergedDistributions:
    '''Merge many distribution election results into one.

    Aggregates distribution election results from a list or dictionary of
    partial results (e.g. by constituency) into a single candidate-wise
    dictionary.
    Neither reconciles ties nor preserves result ordering.

    Use :class:`VoteTotals` to aggregate votes.
    '''
    def convert(self, elected: Union[
                    Dict[Constituency, Dict[Candidate, int]],
                    List[Dict[Candidate, int]]
                ]) -> Dict[Candidate, int]:
        '''Aggregate many distribution election results into one.

        :param elected: Partial distribution election results in a list or
            dictionary; if a dictionary, its keys are ignored and values
            treated as a list.
        '''
        # TODO: reconcile ties when merging? keep result ordering?
        if hasattr(elected, 'values'):
            elected = elected.values()
        merged = {}
        for cdict in elected:
            votelib.util.add_dict_to_dict(merged, cdict)
        return merged


# Constituency-defined vote aggregators by party or district
@simple_serialization
class VoteTotals:
    '''Count total votes/results for each candidate in all constituencies.

    If votes of other type than simple are given, counts the totals of
    votes. Use :class:`MergedDistributions` to aggregate distribution election
    results.
    '''
    def convert(self,
                votes: Dict[Constituency, Dict[Any, int]],
                ) -> Dict[Any, int]:
        '''Count total votes for each candidate in all constituencies.

        :param votes: Votes of any type.
        '''
        all_districts = {}
        for dvotes in votes.values():
            votelib.util.add_dict_to_dict(all_districts, dvotes)
        return all_districts


@simple_serialization
class ConstituencyTotals:
    '''Count total votes (or results) for all candidates in each constituency.

    Accepts any type of votes or distribution election results.

    Useful for apportionment of seats to districts.
    '''
    def convert(self,
                votes: Dict[Constituency, Dict[Any, int]],
                ) -> Dict[Constituency, int]:
        '''Return total votes for all candidates in each constituency.

        :param votes: Votes of any type, or distribution election results.
        '''
        return {
            district: sum(dvotes.values())
            for district, dvotes in votes.items()
        }


@simple_serialization
class PartyTotals:
    '''Count total votes for parties from grouped votes for its candidates.

    Accepts any type of votes or distribution election results.

    Useful for evaluating party-based results in systems where votes can be
    received by individual candidates (panachage) if the results are already
    grouped by party in a nested dictionary, e.g. after applying
    :class:`GroupVotesByParty`.
    '''
    def convert(self,
                votes: Dict[ElectionParty, Dict[Any, int]],
                ) -> Dict[ElectionParty, int]:
        '''Return total votes for all candidates of each party.

        :param votes: Votes of any type, or distribution election results.
        '''
        return {
            party: sum(pvotes.values())
            for party, pvotes in votes.items()
        }


@simple_serialization
class ByConstituency:
    '''Perform conversion for each constituency votes/results separately.

    :param converter: A converter to wrap. Will be called to convert votes (or
        results) for each constituency separately.
    '''
    def __init__(self, converter: Converter):
        self.converter = converter

    def convert(self,
                values: Dict[Constituency, Dict[Any, Any]],
                ) -> Dict[Constituency, Dict[Any, Any]]:
        '''Convert the votes/results for each constituency.

        :param values: Mapping of constituencies to votes or results. The
            keys of this dictionary will be transferred unchanged to the
            result, with their values converted by the wrapped converter.
        '''
        return {
            district: self.converter.convert(dvalues)
            for district, dvalues in values.items()
        }


@simple_serialization
class InvalidVoteEliminator:
    '''Only allow through votes that are declared valid by a given validator.

    Calls the `validate()` method of the validator for each of the keys of the
    input dictionary; if an :class:`vote.VoteError` is raised, does not
    include the vote in the output.

    :param validator: The vote validator to use. Look for some in the
        :mod:`vote` module.
    '''
    def __init__(self, validator: votelib.vote.VoteValidator):
        self.validator = validator

    def convert(self, votes: Dict[Any, int]) -> Dict[Any, int]:
        '''Copy the votes dictionary, removing invalid votes.

        If no invalid votes are detected, does not make a copy.
        '''
        to_remove = []
        for one_vote in votes.keys():
            try:
                self.validator.validate(one_vote)
            except votelib.vote.VoteError:
                to_remove.append(one_vote)
        if to_remove:
            votes = votes.copy()
            for one_vote in to_remove:
                del votes[one_vote]
        return votes


@simple_serialization
class RoundedVotes:
    '''Round vote counts to the given number of decimal digits.

    In some systems, rounding of fractional values to a given number of digits
    is specified (e.g. Scottish STV). This rounds vote counts of any vote type
    to this number of decimals.

    :param decimals: Number of digits to round to. Negative values are not
        accepted.
    :param round_method: A rounding method accepted by Python's ``decimal``
        module.
    :raises ValueError: If an invalid number of decimal digits is given.
    '''
    def __init__(self, decimals: int, round_method=decimal.ROUND_HALF_UP):
        self.decimals = decimals
        self.round_method = round_method
        self._rounder = self._construct_rounder()

    def _exponent(self):
        if self.decimals < 0:
            raise ValueError(f'invalid number of decimals: {self.decimals}')
        elif self.decimals == 0:
            return Decimal('1')
        elif self.decimals == 1:
            return Decimal('.1')
        else:
            return Decimal('.' + ('0' * (self.decimals - 1)) + '1')

    def _construct_rounder(self):
        decprec = self._exponent()

        # using default arguments to speed up lookups
        def _rounder(val, round_prec=decprec, round_method=self.round_method):
            if isinstance(val, Fraction):
                decimal_val = Decimal(val.numerator) / Decimal(val.denominator)
            else:
                decimal_val = Decimal(val)
            return decimal_val.quantize(round_prec, round_method)

        return _rounder

    def convert(self, votes: Dict[Any, Number]) -> Dict[Any, Decimal]:
        '''Round all vote counts to the specified number of votes.

        The vote counts must be convertible to :class:`Decimal`
        (:class:`Fraction` and any types accepted by the decimal constructor
        are supported).

        :param votes: Votes whose counts should be rounded.
        '''
        return {
            vote: self._rounder(n_votes) for vote, n_votes in votes.items()
        }


@simple_serialization
class SubsettedVotes:
    '''Subset the votes to only concern a subset of candidates.

    A wrapper over :class:`VoteSubsetter` that takes the entire vote count
    dictionary, not just a single vote object (key).
    Useful when some candidates should be excluded because they do not pass
    an electoral threshold, or when evaluating ties.

    :param vote_subsetter: A vote subsetter that turns a single vote object
        (key) into a vote object that only concerns the specified candidates,
        with other candidates removed.
    '''
    DEFAULT_SUBSETTER = votelib.vote.SimpleSubsetter()

    def __init__(self,
                 vote_subsetter: VoteSubsetter = DEFAULT_SUBSETTER,
                 depth: int = 0,
                 ):
        self.vote_subsetter = vote_subsetter
        self.depth = depth

    def convert(self,
                votes: Dict[Any, Number],
                subset: Collection[Candidate],
                ) -> Dict[Any, Number]:
        '''Subset the votes to only concern a subset of candidates.

        :param votes: Votes to be subsetted. Their type should be in accordance
            with the wrapped vote subsetter.
        :param subset: The only candidates that should be contained in the
            output.
        '''
        return self._convert(votes, subset, depth=self.depth)

    def _convert(self,
                 votes: Dict[Any, Number],
                 subset: Collection[Candidate],
                 depth: int,
                 ) -> Dict[Any, Number]:
        if depth == 0:
            sub = collections.defaultdict(int)
            for full_vote, n_votes in votes.items():
                sub_vote = self.vote_subsetter.subset(full_vote, subset)
                if sub_vote is not None:
                    sub[sub_vote] += n_votes
            return dict(sub)
        else:
            return {
                nester: self._convert(nested, subset, depth - 1)
                for nester, nested in votes.items()
            }


RANKED_TO_SIMPLE: List[type] = [
    RankedToPositionalVotes,
    RankedToFirstPreference,
    RankedToPresenceCounts,
]


@simple_serialization
class Chain:
    """Chain multiple vote converters after one another.

    Applies the converters successively on a single vote dictionary.
    """
    def __init__(self, converters: List[Converter]):
        self.converters = converters

    def convert(self, votes: Dict[Any, Number]) -> Dict[Any, Number]:
        output = votes
        for conv in self.converters:
            output = conv.convert(output)
        return output



'''Objects to assign scores to ranks in ranked voting systems such as Borda.

A rank scorer returns a list of numerical scores to be assigned to ranks given
by voters. This is the essence of Borda count system, and rank scorers capture
most of the variations there are in that system.
'''

import abc
from fractions import Fraction
from typing import List, Any
from numbers import Number

from votelib.persist import simple_serialization


def select_padded(sequence: List[Any], n: int, pad_with: Any = 0) -> List[Any]:
    '''Select n leading elements from sequence, padding with pad_with.

    Padding with pad_with is used when sequence is not long enough to select
    n elements.
    '''
    selected = sequence[:n]
    if n > len(selected):
        selected += [pad_with] * (n - len(selected))
    return selected


class RankScorer(metaclass=abc.ABCMeta):
    '''An abstract base class for rank scorers.

    Rank scorers must provide a `scores()` method that returns a list of scores
    based on the number of ranks given. They may also provide a
    `set_n_candidates()` method that sets the total number of candidates
    participating in the election, which might be relevant for computing the
    rank scores. If this method is defined, it must be called before the
    `scores()` method is called first.
    '''
    @abc.abstractmethod
    def scores(self, n_ranked: int) -> List[Number]:
        raise NotImplementedError


@simple_serialization
class Borda(RankScorer):
    '''Borda rank scorer, corresponding to the original Borda count variant.

    Assigns the `base` score to the candidate ranked last, and one point more
    for each higher rank.

    This rank scorer needs to be initialized by the :func:`set_n_candidates()`
    before calling :func:`scores()`.

    :param base: The score to assign to the candidate ranked last. For the
        truly original Borda, this equals to 1; some variants set it to zero,
        and thus set the score for the first rank to the number of candidates
        minus one.
    '''

    def __init__(self, base: int = 1):
        self.base = base
        self._scores = None
        self.n_candidates = None

    def set_n_candidates(self, n_candidates: int) -> None:
        '''Set the total number of candidates that could be ranked.

        This helps to account for rankings that do not rank all candidates.

        :param n_candidates: The total number of candidates that could be
            ranked on any ballot (i.e. the number of candidates participating
            in the election in the particular constituency).
        '''
        self.n_candidates = n_candidates
        top_score = self.n_candidates + self.base - 1
        self._scores = [
            top_score - rank
            for rank in range(self.n_candidates)
        ]

    def get_n_candidates(self) -> int:
        return self.n_candidates

    def scores(self, n_ranked: int) -> List[int]:
        '''Return the scores for the first n_ranked ranks.

        This gives (number of candidates + base - 1 - rank) for ranks running
        from 0 (best rank) to n_ranked.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        :raises RuntimeError: If the scorer has not been initialized first by
            calling ``set_n_candidates()``.
        '''
        try:
            if n_ranked > self.n_candidates:
                raise ValueError(f'cannot rank {n_ranked} out of maximum'
                                 f' {self.n_candidates} candidates')
            else:
                return select_padded(self._scores, n_ranked)
        except TypeError:
            raise RuntimeError(
                'scorer not initialized, call set_n_candidates() first'
            )


@simple_serialization
class Dowdall(RankScorer):
    '''Dowdall (Nauru) rank scorer.

    Assigns the numbers of the harmonic series (1, 1/2, 1/3...) to
    progressively lower ranks.
    '''

    def scores(self, n_ranked: int) -> List[Fraction]:
        '''Return the scores for the first n_ranked ranks.

        This gives `1 / (rank + 1)` for ranks running
        from 0 (best rank) to n_ranked.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        '''
        return [Fraction(1, rank + 1) for rank in range(n_ranked)]


@simple_serialization
class Geometric(RankScorer):
    '''A geometric progression rank scorer.

    Assigns the numbers of a chosen inverse geometric progression
    (e.g. 1, 1/2, 1/4... for 2) to progressively lower ranks.

    :param base: Base of the geometric progression.
    '''
    # http://www.geometric-voting.org.uk/index.htm

    def __init__(self, base: int = 2):
        self.base = base

    def scores(self, n_ranked: int) -> List[Fraction]:
        '''Return the scores for the first n_ranked ranks.

        This gives `1 / (2 ** rank)` for ranks running
        from 0 (best rank) to n_ranked.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        '''
        return [Fraction(1, self.base ** rank) for rank in range(n_ranked)]


@simple_serialization
class ModifiedBorda(RankScorer):
    '''Modified Borda count rank scorer.

    In this system, the score for the highest rank is not constant (as is the
    case for the vanilla Borda count), but is equal to the number of ranked
    candidates; therefore, it encourages voters to rank many candidates.
    '''

    def scores(self, n_ranked: int) -> List[int]:
        '''Return the scores for the first n_ranked ranks.

        This gives `n_ranked - rank` for ranks running
        from 0 (best rank) to n_ranked.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        '''
        return [n_ranked - rank for rank in range(n_ranked)]


@simple_serialization
class FixedTop(RankScorer):
    '''A rank scorer with fixed score for the top rank.

    Assigns scores progressively decreased by one until hitting zero.

    :param top: The score for the top (best) ranked candidate on the ballot.
    '''

    def __init__(self, top: int):
        self.top = top

    def scores(self, n_ranked: int) -> List[int]:
        '''Return the scores for the first n_ranked ranks.

        This gives `max(top - rank, 0)` for ranks running
        from 0 (best rank) to n_ranked.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        '''
        return [max(self.top - rank, 0) for rank in range(n_ranked)]


@simple_serialization
class SequenceBased(RankScorer):
    '''A rank scorer with a predetermined sequence of scores.

    This is used in many competitions (Eurovision, Formula One, etc.)

    Assigns scores according to the given sequence until hitting zero, and zero
    scores afterwards.

    :param sequence: The scores for the top candidates on the ballot.
    '''

    def __init__(self, sequence: List[Number]):
        self.sequence = sequence

    def scores(self, n_ranked: int) -> List[Number]:
        '''Return the scores for the first n_ranked ranks.

        This gives values from the initial sequence, then zeros.

        :param n_ranked: Number of ranks to be returned. Equal to the length
            of the output list.
        '''
        return select_padded(self.sequence, n_ranked)



'''Functions to score magnitudes of wins between pairs of candidates.

These are used in some Condorcet methods to determine ranking priority.
'''

from typing import Callable, Dict, Tuple
from numbers import Number

import votelib.component.core
from votelib.vote import Candidate


PAIRWIN_SCORERS = {}


pairwin_scorer_mark, get, construct = \
    votelib.component.core.register_functions(
        PAIRWIN_SCORERS, 'pairwise win scorer', Callable[
            [Dict[Tuple[Candidate, Candidate], Number]],
            Dict[Tuple[Candidate, Candidate], Number]
        ]
    )


@pairwin_scorer_mark
def winning_votes(counts: Dict[Tuple[Candidate, Candidate], Number]
                  ) -> Dict[Tuple[Candidate, Candidate], Number]:
    '''Winning votes pairwise win scorer. Counts wins fully, zero otherwise.

    This is the most common pairwise win scorer. When the number of votes
    for the pair ranked in one direction is larger than the other direction,
    assigns all those votes as the pairwise win strength.

    :param counts: Condorcet votes (counts of pairwise preferences).
    '''
    return {
        pair: (count if count > counts.get(tuple(reversed(pair)), 0) else 0)
        for pair, count in counts.items()
    }


@pairwin_scorer_mark
def margins(counts: Dict[Tuple[Candidate, Candidate], Number]
            ) -> Dict[Tuple[Candidate, Candidate], Number]:
    '''Margins pairwise win scorer. Takes the difference from reverse option.

    Also called margin of victory or defeat strength. Assigns the number of
    votes ranking the pair in the given order minus the number of votes doing
    the reverse as the win strength (which is thus negative for pairwise
    losses).

    :param counts: Condorcet votes (counts of pairwise preferences).
    '''
    return {
        pair: count - counts.get(tuple(reversed(pair)), 0)
        for pair, count in counts.items()
    }


@pairwin_scorer_mark
def pairwise_opposition(counts: Dict[Tuple[Candidate, Candidate], Number]
                        ) -> Dict[Tuple[Candidate, Candidate], Number]:
    '''Pairwise opposition win scorer. Returns the win counts unchanged.

    This gives the number of votes ranking the pair in the given order
    directly as the measure of pairwise win, regardless of the number of votes
    preferring the opposite pairwise ranking.

    :param counts: Condorcet votes (counts of pairwise preferences).
    '''
    return counts



"""Create and plot Yee diagrams outlining single-winner voting system quality.

Yee diagrams distribute some candidates in a 2D issue space and then evaluate
a selected voting system on a grid over that space. In each cell of the grid,
an election is simulated with voter preferences proportional to the distances
of the grid cell to the candidate positions. Use :func:`diagram` to produce
this.

A good voting system's Yee diagram should closely approximate a Voronoi diagram
over the candidates. (A Voronoi diagram would assign each point of space to the
closest candidate.) [#yee]_ This can be produced by calling :func:`voronoi`.

The Yee diagram here is represented by a 2D list as a rectangular lattice
covering the 0-1 in both dimensions.

.. [#yee] Warren D. Smith. "Yee Pictures", Range Voting, 2007.
    https://rangevoting.org/IEVS/Pictures.html
"""

import math
import random
from typing import Tuple, List, Dict, Optional

try:
    import matplotlib.cm
    import matplotlib.colors
    import matplotlib.patches
    import matplotlib.pyplot as plt
except ImportError:
    matplotlib = None
    plt = None

import votelib.convert
import votelib.evaluate
import votelib.evaluate.auxiliary
import votelib.evaluate.cardinal
import votelib.generate
import votelib.vote
from votelib.candidate import Candidate


DEFAULT_SHAPE = (200, 200)


def diagram(evaluator: votelib.evaluate.Selector,
            candidates: Dict[Candidate, Tuple[float, float]],
            shape: Tuple[int, int] = DEFAULT_SHAPE,
            n_votes: int = 1000,
            random_state: Optional[int] = None,
            **kwargs) -> List[List[Candidate]]:
    """Produce a Yee diagram for the evaluator.

    Any superfluous keyword arguments are passed on to the vote generation
    sampler using :func:`sampler`.

    :param evaluator: The voting system evaluator for which to construct
        a Yee diagram.
    :param candidates: Candidate positions in the 2D issue space. The
        coordinates should both be between 0 and 1.
    :param shape: Number of cells along each dimension of the issue space.
        Enlarging this gives more detail but increases computing time.
    :param n_votes: Number of voters to consider in each simulated election.
        Lowering this will speed up the computation but may give unstable
        results.
    :param random_state: Seed for the vote generator.
    """
    # Ensure that we always have a decisive winner by breaking ties randomly.
    full_eval = votelib.evaluate.TieBreaking(
        evaluator,
        tiebreaker=votelib.evaluate.PreConverted(
            votelib.convert.ScoreToSimpleVotes(),
            votelib.evaluate.auxiliary.Sortitor(),
        ),
        subsetter=votelib.vote.ScoreSubsetter(),
    )
    if random_state is not None:
        random.seed(random_state)
    diag = []
    ys, xs = _diag_coors(shape)
    for y in ys:
        row = []
        for x in xs:
            votes = votelib.generate.IssueSpaceGenerator(
                candidates,
                sampler=sampler(x, y, **kwargs),
            ).generate(n_votes)
            result = full_eval.evaluate(votes, n_seats=1)
            row.append(result[0])
        diag.append(row)
    return diag


def _diag_coors(shape: Tuple[int, int]) -> Tuple[List[float], List[float]]:
    y_size_elem, x_size_elem = [1 / s for s in shape]
    return (
        [(i + .5) * y_size_elem for i in range(shape[0])],
        [(i + .5) * x_size_elem for i in range(shape[1])]
    )


def sampler(x: float,
            y: float,
            sigma: Tuple[float, float] = (.25, .25),
            ) -> votelib.generate.BoundedSampler:
    """Create a vote generation sampler for a given Yee issue space point.

    Gives a Gaussian sampler centered on the x-y point with a standard
    deviation given by sigma.
    """
    return bounded_sampler(votelib.generate.DistributionSampler(
        distribution='gauss', mu=(x, y), sigma=sigma
    ))


def bounded_sampler(samp: votelib.generate.Sampler
                    ) -> votelib.generate.BoundedSampler:
    """Bound a vote sampler to the Yee diagram bounding box."""
    return votelib.generate.BoundedSampler(samp, (0, 0, 1, 1))


def voronoi(candidates: Dict[Candidate, Tuple[float, float]],
            shape: Tuple[int, int] = DEFAULT_SHAPE,
            ) -> List[List[Candidate]]:
    """Produce a Voronoi Yee diagram for the given candidates.

    This gives the ideal Yee diagram that good voting systems should be
    close to.

    :param candidates: Candidate positions in the 2D issue space. The
        coordinates should both be between 0 and 1.
    :param shape: Number of cells along each dimension of the issue space.
        Enlarging this gives more detail but increases computing time.
    """
    ys, xs = _diag_coors(shape)
    return [
        [min(
            candidates.items(),
            key=lambda item: math.hypot(item[1][0] - x, item[1][1] - y)
        )[0] for x in xs]
        for y in ys
    ]


def voronoi_matches(results: List[List[Candidate]],
                    candidates: Dict[Candidate, Tuple[float, float]],
                    ) -> List[List[bool]]:
    """Give a 2D boolean mask how a Yee diagram matches a Voronoi diagram.

    :param results: A Yee diagram of winning candidates.
    :param candidates: Candidate positions in the 2D issue space. The
        coordinates should both be between 0 and 1.
    """
    shape = (len(results), len(results[0]))
    voronoi_diagram = voronoi(candidates, shape=shape)
    return [
        [results[i][j] == voronoi_diagram[i][j] for j in range(shape[1])]
        for i in range(shape[0])
    ]


def voronoi_conformity(results: List[List[Candidate]],
                       candidates: Dict[Candidate, Tuple[float, float]],
                       ) -> float:
    """Calculate the percentage of how a Yee diagram matches a Voronoi diagram.

    :param results: A Yee diagram of winning candidates.
    :param candidates: Candidate positions in the 2D issue space. The
        coordinates should both be between 0 and 1.
    """
    shape = (len(results), len(results[0]))
    matches = voronoi_matches(results, candidates)
    return sum(m for row in matches for m in row) / (shape[0] * shape[1])


def circular_candidates(n: int,
                        r: float = .25,
                        center: Tuple[float, float] = (.5, .5),
                        ) -> Dict[Candidate, Tuple[float, float]]:
    """Generate Yee diagram candidate issue space positions in a circle.

    :param n: Number of candidates. The candidates will be evenly spaced
        along the circumference of the circle.
    :param r: Radius of the circle.
    :param center: Center of the circle coordinates.
    """
    unit_angle = 2 * math.pi / n
    cx, cy = center
    return dict(zip(votelib.generate.candidate_names(n), (
        (
            cx + r * math.cos(unit_angle * i),
            cy + r * math.sin(unit_angle * i)
        ) for i in range(n)
    )))


def random_candidates(n: int,
                      cand_sampler: Optional[votelib.generate.Sampler] = None,
                      ) -> Dict[Candidate, Tuple[float, ...]]:
    """Generate Yee diagram candidate issue space positions randomly.

    :param n: Number of candidates.
    :param cand_sampler: A sampler to generate the candidate positions.
        The default (None) is to sample from a uniform 2D distribution, i.e.
        all points of the diagram are equally probable.
    """
    if cand_sampler is None:
        cand_sampler = votelib.generate.DistributionSampler(
            'uniform', n_dims=2
        )
    cand_sampler = bounded_sampler(cand_sampler)
    return dict(zip(
        votelib.generate.candidate_names(n),
        cand_sampler.sample(n)
    ))


def plot(results: List[List[Candidate]],
         candidates: Dict[Candidate, Tuple[float, float]],
         cmap: str = 'tab10',
         ax=None) -> None:
    """Plot the Yee diagram.

    Requires Matplotlib. Plots the Yee diagram cells and candidate positions
    on the current/chosen axes and adds a legend. You normally need to follow
    up on this with the ``show()`` call to show the figure.

    :param results: The Yee diagram to plot.
    :param candidates: Candidate positions in the 2D issue space. The
        coordinates should both be between 0 and 1.
    :param cmap: Colormap to use for plotting. The default is sufficient
        for up to 10 candidates unambiguously. References the Matplotlib
        colormap register.
    :param ax: Axes to plot on.
    """
    if plt is None:
        raise ImportError('votelib.crit.yee.plot requires matplotlib')
    if ax is None:
        ax = plt.gca()
    if isinstance(cmap, str):
        cmap = matplotlib.cm.get_cmap(cmap)
    # Convert candidate objects in the diagram to integers to plot easily.
    cand_list = list(sorted(candidates.keys()))
    cand_coors = [candidates[c] for c in cand_list]
    cand_x, cand_y = zip(*cand_coors)
    results_int = [[cand_list.index(cand) for cand in row] for row in results]
    norm = matplotlib.colors.Normalize(vmin=0, vmax=len(cmap.colors)-1)
    ax.imshow(
        results_int,
        cmap=cmap,
        extent=(0, 1, 0, 1),
        interpolation='none',
        resample=False,
        norm=norm,
        alpha=.75,
        origin='lower',
    )
    ax.scatter(
        cand_x, cand_y,
        c=range(len(cand_list)),
        cmap=cmap,
        norm=norm,
        edgecolors='white',
    )
    legend_handles = [
        matplotlib.patches.Patch(color=cmap(i), label=str(cand))
        for i, cand in enumerate(cand_list)
    ]
    ax.legend(handles=legend_handles)



'''Objects to transfer votes between candidates for transferable vote systems.

Transfers of votes from eliminated and elected candidates to candidates staying
in the contest are an essential part of any transferable vote system,
such as the one of
:class:`votelib.evaluate.sequential.TransferableVoteSelector`.
There are many variants how to achieve this, some are implemented here.

NOTE: The :class:`VoteTransferer` interface is provisional and may be changed
in the future, chiefly to accommodate a wider range of transfer methods (Meek,
Wright...)
'''

import sys
import abc
import collections
import random
import itertools
import bisect
import functools
from fractions import Fraction
from typing import Any, List, Dict, FrozenSet, Optional, Collection, Union
from numbers import Number

from votelib.candidate import Candidate
from votelib.vote import RankedVoteType
from votelib.evaluate.proportional import LargestRemainder
from votelib.persist import simple_serialization


RankedVoteAllocation = Dict[Optional[Candidate], Dict[RankedVoteType, Number]]


def ranked_next(vote: RankedVoteType,
                cand: Candidate,
                allowed: Collection[Candidate],
                ) -> FrozenSet[Candidate]:
    '''Select the candidate(s) ranked in the vote after the given candidate.

    :param vote: The ranked vote to examine.
    :param cand: The candidate to look after.
    :param allowed: Only return candidate(s) from this collection, skip all
        others to lower ranks.
    :returns: Candidates ranked after cand. Will be empty if cand was ranked
        last or is not present in the vote. Will only have multiple members
        if the ranked vote contains a shared rank after cand.
    '''
    # Iterate through the vote until we find cand. At him, set take_next to
    # True to return the candidates ranked next.
    take_next = False
    for rank_alt in vote:
        if isinstance(rank_alt, frozenset):
            if take_next:
                allowed_alt = rank_alt.intersection(allowed)
                if allowed_alt:
                    return allowed_alt
                # else go on for another rank
            elif cand in rank_alt:
                take_next = True
        else:
            if take_next:
                if rank_alt in allowed:
                    return frozenset([rank_alt])
            elif cand == rank_alt:
                take_next = True
    return frozenset()    # exhausted ballot


def distribute_n_random(cand_weights: Dict[Any, Number],
                        n: int,
                        limit_by_weight: bool = False,
                        ) -> Dict[Any, int]:
    '''Distribute n randomly among candidates with weighted probabilities.

    :param cand_weights: Candidates and their weights.
    :param n: Number to distribute.
    :param limit_by_weight: Whether the maximum count assigned to a candidate
        is limited by the value of their weight.
    :returns: A dictionary mapping candidates to their assigned quantities.
        The quantities sum to n.
    '''
    candidates, weights = zip(*cand_weights.items())
    total_weight = sum(weights)
    if isinstance(total_weight, Fraction):
        max_denom = max(
            f.denominator if hasattr(f, 'denominator') else 1
            for f in weights
        )
        # multiply weights by maximum denominator and then go on with integers
        weights = [w * max_denom for w in weights]
        total_weight *= max_denom
    if isinstance(total_weight, int):
        cum_weights = list(itertools.accumulate(weights))
        generator = random.sample(range(total_weight), n)
    else:
        cum_weights = list(itertools.accumulate(
            w / total_weight for w in weights
        ))
        generator = (random.random() for i in range(n))
    do_bisect = functools.partial(bisect.bisect_right, cum_weights)
    index_selection = collections.Counter(
        do_bisect(i) for i in generator
    )
    selection = {candidates[ind]: n for ind, n in index_selection.items()}
    if limit_by_weight:
        return LargestRemainder('hare').evaluate(
            selection, n, max_seats=cand_weights
        )
    else:
        return selection


class VoteTransferer(metaclass=abc.ABCMeta):
    '''An abstract base class for vote transferers.

    Vote transferers must provide a `subtract()` method that subtracts votes
    from elected candidates and a `transfer()` method that redistributes
    votes for some candidates to those remaining in the contest.
    '''
    @abc.abstractmethod
    def subtract(self,
                 allocation: RankedVoteAllocation,
                 elected: Dict[Candidate, int],
                 ) -> RankedVoteAllocation:
        '''Remove votes from elected candidates according to the quota.

        :param allocation: Current allocation of ranked votes to candidates.
            The votes allocated to elected candidates will be lowered by the
            amount corresponding to the quota for their election.
        :param elected: Elected candidates, mapped to the quota with which they
            were elected (or multiples thereof, if they were awarded multiple
            seats). These quotas should be removed from the candidates'
            votes (because so many votes were used).
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def transfer(self,
                 allocation: RankedVoteAllocation,
                 candidates: List[Candidate],
                 ) -> RankedVoteAllocation:
        '''Transfer votes from eliminated or fully elected candidates.

        :param allocation: Current allocation of ranked votes to candidates.
            The votes allocated to specified candidates will be reallocated
            to other candidates.
        :param candidates: Candidates no longer continuing in the contest.
        '''
        raise NotImplementedError


class SimpleVoteTransferer(VoteTransferer):
    def subtract(self,
                 allocation: RankedVoteAllocation,
                 elected: Dict[Candidate, int],
                 ) -> RankedVoteAllocation:
        allocation = {cand: alloc.copy() for cand, alloc in allocation.items()}
        if elected:
            for cand, quota in elected.items():
                self._subtract(allocation[cand], quota)
        return allocation

    def transfer(self,
                 allocation: RankedVoteAllocation,
                 candidates: List[Candidate],
                 ) -> RankedVoteAllocation:
        '''Transfer votes from eliminated or fully elected candidates.

        :param allocation: Current allocation of ranked votes to candidates.
            The votes allocated to specified candidates will be reallocated
            to other candidates that are present in the allocation, or
            reassigned to the None key as exhausted ballots.
        :param candidates: Candidates to be removed from the allocation.
        '''
        allocation = {cand: alloc.copy() for cand, alloc in allocation.items()}
        to_remove = [cand for cand in allocation if cand in candidates]
        continuing = [
            cand for cand in allocation
            if cand is not None and cand not in candidates
        ]
        for cand in to_remove:
            for vote, n_votes in allocation[cand].items():
                targets = ranked_next(vote, cand, continuing)
                if targets:
                    if len(targets) > 1:
                        realloc = self._distribute_equal_ranking(
                            targets, n_votes
                        )
                    else:
                        target, = targets
                        realloc = {target: n_votes}
                    for target, n in realloc.items():
                        target_alloc = allocation.setdefault(target, {})
                        if vote not in target_alloc:
                            target_alloc[vote] = 0
                        target_alloc[vote] += n
                else:
                    # Put the ballots on the exhausted pile.
                    if None not in allocation:
                        allocation[None] = {}
                    if vote not in allocation[None]:
                        allocation[None][vote] = 0
                    allocation[None][vote] += n_votes
            del allocation[cand]
        return allocation

    def _subtract(self,
                  cand_alloc: Dict[RankedVoteType, Number],
                  quota: int,
                  ) -> None:
        # Remove a total of *quota* votes from values of *cand_alloc*.
        # Hare STV does this randomly, Gregory STV uses exact proportional
        # fractions.
        raise NotImplementedError

    def _distribute_equal_ranking(self,
                                  targets: FrozenSet[Candidate],
                                  n_votes: int,
                                  ) -> Dict[Candidate, int]:
        # Distribute a total of *n_votes* among candidates in *targets*.
        # Hare STV does this randomly, Gregory STV uses exact proportional
        # fractions.
        raise NotImplementedError


@simple_serialization
class Hare(SimpleVoteTransferer):
    '''Hare (random ballot selection) vote transferer.

    This is the variant used (AFAIK) in Irish lower house legislative elections
    (Dáil Éireann).

    When a candidate is elected by quota, the number of ballots corresponding
    to the quota is randomly selected and discarded, the rest is fully
    transferred to their next preferences.

    In case of shared ranks, the votes are randomly distributed between the
    candidates sharing the rank.

    :param seed: Seed for the random generator.
    '''
    def __init__(self,
                 seed: Optional[int] = None,
                 ):
        self.seed = seed
        self.stable = (self.seed is not None)

    def _subtract(self,
                  cand_alloc: Dict[RankedVoteType, Number],
                  n_sub: int,
                  ) -> None:
        random.seed(self.seed)
        subtractions = distribute_n_random(
            cand_alloc, n_sub, limit_by_weight=True
        )
        for vote, to_sub in subtractions.items():
            to_sub_from = cand_alloc[vote]
            if to_sub >= to_sub_from:
                del cand_alloc[vote]
            else:
                cand_alloc[vote] -= to_sub

    def _distribute_equal_ranking(self,
                                  targets: FrozenSet[Candidate],
                                  n_votes: int,
                                  ) -> Dict[Candidate, int]:
        whole_val = n_votes // len(targets)
        if whole_val:
            result = {cand: whole_val for cand in targets}
            remainder = n_votes - len(targets) * whole_val
            if not remainder:
                return result
        else:
            result = {}
            remainder = n_votes
        random.seed(self.seed)
        remnant_dist = distribute_n_random(
            {tgt: remainder for tgt in targets},
            remainder,
        )
        for cand, transfer in remnant_dist.items():
            result[cand] = result.get(cand, 0) + transfer
        return result


@simple_serialization
class Gregory(SimpleVoteTransferer):
    '''Gregory (fractional) vote transferer.

    The variant used is the Weighted Inclusive Gregory Method (WIGM) used e.g.
    in Scottish local government elections.

    When a candidate is elected by quota, the fraction corresponding to the
    quota divided by total votes for the candidate is used to multiply (lower)
    each vote allocated to that candidate; the votes are then transferred to
    next candidates on the ballots.

    In case of shared ranks, the votes are evenly distributed between the
    candidates sharing the rank.

    The implementation produces exact fractional votes. Rounding rules are not
    implemented yet.
    '''
    def _subtract(self,
                  cand_alloc: Dict[RankedVoteType, Fraction],
                  n_sub: Fraction,
                  ) -> None:
        current_sum = sum(cand_alloc.values())
        if current_sum == 0:
            raise RuntimeError
        if n_sub >= current_sum:
            # We need to subtract more than there is, delete all votes.
            for vote in list(cand_alloc.keys()):
                del cand_alloc[vote]
        else:
            fraction = Fraction(current_sum - n_sub, current_sum)
            for vote in cand_alloc.keys():
                cand_alloc[vote] *= fraction

    def _distribute_equal_ranking(self,
                                  targets: FrozenSet[Candidate],
                                  n_votes: Fraction,
                                  ) -> Dict[Candidate, Fraction]:
        split_val = Fraction(n_votes, len(targets))
        return {cand: split_val for cand in targets}


def construct(trans_def: Union[VoteTransferer, str]) -> VoteTransferer:
    return (
        getattr(sys.modules[__name__], trans_def)()
        if isinstance(trans_def, str)
        else trans_def
    )



'''Quota functions used in largest-remainder proportional voting systems.

Quotas can however be used as a component in many other voting systems,
such as transferable vote or open list evaluators.

A quota function takes the total number of votes and the number of seats
to allocate and returns the number of votes required to reach a seat.
The unrounded quota functions return fractions to retain exact values.

All supported quota functions are assembled in the `QUOTAS` dictionary keyed
by their name. `get()` retrieves from this dictionary by string key;
`construct()` also accepts callables and passes them through.
'''

import math
from fractions import Fraction
from typing import Callable
from numbers import Number

import votelib.component.core


QUOTAS = {}


quota_mark, get, construct = votelib.component.core.register_functions(
    QUOTAS, 'quota', Callable[[int, int], Number]
)


@quota_mark
def hare(votes: int, seats: int) -> Fraction:
    '''Hare quota, the most basic one.

    This is the unrounded variant, giving the exact fraction.
    '''
    return Fraction(votes, seats)


@quota_mark
def hare_rounded(votes: int, seats: int) -> int:
    '''Hare quota, the most basic one.

    This is the rounded variant, which is used more often. Half is rounded up.
    '''
    return int(_round_half_up(Fraction(votes, seats)))


@quota_mark
def droop(votes: int, seats: int) -> int:
    '''Droop quota, the most widely used one.

    This it is the smallest integer quota guaranteeing the number of passing
    candidates will not be higher than the number of seats.
    '''
    return int(Fraction(votes, seats + 1)) + 1


@quota_mark
def hagenbach_bischoff(votes: int, seats: int) -> Fraction:
    '''Hagenbach-Bischoff quota.

    This is the unrounded variant, giving the exact fraction.
    '''
    return Fraction(votes, seats + 1)


@quota_mark
def hagenbach_bischoff_ceil(votes: int, seats: int) -> int:
    '''Hagenbach-Bischoff quota, rounding up.

    This is the rounded variant that is identical to the Droop quota in most
    cases.
    '''
    return int(math.ceil(Fraction(votes, seats + 1)))


@quota_mark
def hagenbach_bischoff_rounded(votes: int, seats: int) -> int:
    '''Hagenbach-Bischoff quota, rounding mathematically (round-to-even).

    This is the rounded variant that is used in some rare cases, e.g. old
    Slovak regional apportionment of parliamentary seats. Half is rounded up.
    '''
    return int(_round_half_up(Fraction(votes, seats + 1)))


@quota_mark
def imperiali(votes: int, seats: int) -> Fraction:
    '''Imperiali quota.

    Imperiali quota can produce more candidates than seats to be filled in some
    cases; the results then usually need to be recalculated using a different
    quota.
    '''
    return Fraction(votes, seats + 2)


def _round_half_up(var: Fraction) -> int:
    if var.limit_denominator(2) == var:
        # This concerns halves (rounding up) and integrals (no effect)
        return int(math.ceil(var))
    else:
        return int(round(var))


class constant:
    def __init__(self, quota: Number):
        self.quota = quota

    def __call__(self, votes: int, seats: int) -> Number:
        return self.quota
